use bevy::prelude::*;
use std::time::Duration;

use bevy::utils::HashMap;

use std::hash::Hash;

pub fn run_animations<Tag: Hash + Eq + Component>(
    mut query: Query<(
        &Tag,
        &mut AnimationManager<Tag>,
        &mut TextureAtlas,
        &mut Handle<Image>,
    )>,
    time: Res<Time>,
) {
    for (tag, mut manager, mut atlas, mut tex_handle) in &mut query {
        let anim = manager
            .animations
            .get_mut(tag)
            .expect("animations should exist for all tags!");

        *tex_handle = anim.tex.clone_weak();
        atlas.layout = anim.atlas_layout_handle.clone_weak();
        if atlas.index > anim.max_index {
            atlas.index = 0
        }

        anim.frame_timer.tick(time.delta());

        if anim.frame_timer.just_finished() {
            dbg!(atlas.index);
            if atlas.index >= anim.max_index {
                atlas.index = 0;
            } else {
                atlas.index += 1;
            }
            anim.frame_timer = Animation::timer_from_fps(anim.fps);
        }
    }
}

#[derive(Bundle)]
pub struct AnimBundle<Tag: Hash + Eq + Component> {
    pub tag: Tag,
    pub mgr: AnimationManager<Tag>,
}

#[derive(Default, Component)]
pub(crate) struct AnimationManager<Tag: Hash + Eq> {
    pub animations: HashMap<Tag, Animation>,
}

impl<Tag: Hash + Eq> AnimationManager<Tag> {
    pub fn insert(mut self, which: Tag, anim: Animation) -> Self {
        self.animations.insert(which, anim);
        self
    }
}

pub(crate) struct Animation {
    atlas_layout_handle: Handle<TextureAtlasLayout>,
    max_index: usize,
    fps: u8,
    frame_timer: Timer,
    tex: Handle<Image>,
}

impl Animation {
    pub fn new(
        atlas_layout_handle: Handle<TextureAtlasLayout>,
        length: usize,
        fps: u8,
        tex_handle: Handle<Image>,
    ) -> Self {
        Self {
            atlas_layout_handle,
            max_index: length - 1,
            fps,
            frame_timer: Self::timer_from_fps(fps),
            tex: tex_handle,
        }
    }

    // stolen from https://github.com/bevyengine/bevy/blob/5c759a1be800209f537bea31d32b8ba7e966b0c1/examples/2d/sprite_animation.rs#L53-L55
    pub(crate) fn timer_from_fps(fps: u8) -> Timer {
        Timer::new(
            Duration::from_secs_f32(1.0 / (fps as f32)),
            TimerMode::Repeating,
        )
    }
}
